/*
 * dn.c
 *
 *    This file contains dummynet's utility program.
 *
 *
 *            Copyright (C) 1999 Hewlett-Packard Company
 *                       ALL RIGHTS RESERVED.
 *
 * The enclosed software and documentation includes copyrighted works
 * of Hewlett-Packard Co. For as long as you comply with the following
 * limitations, you are hereby authorized to (i) use, reproduce, and
 * modify the software and documentation, and to (ii) distribute the
 * software and documentation, including modifications, for
 * non-commercial purposes only.
 *      
 * 1.  The enclosed software and documentation is made available at no
 *     charge in order to advance the general development of
 *     high-performance networking products.
 * 
 * 2.  You may not delete any copyright notices contained in the
 *     software or documentation. All hard copies, and copies in
 *     source code or object code form, of the software or
 *     documentation (including modifications) must contain at least
 *     one of the copyright notices.
 * 
 * 3.  The enclosed software and documentation has not been subjected
 *     to testing and quality control and is not a Hewlett-Packard Co.
 *     product. At a future time, Hewlett-Packard Co. may or may not
 *     offer a version of the software and documentation as a product.
 * 
 * 4.  THE SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS".
 *     HEWLETT-PACKARD COMPANY DOES NOT WARRANT THAT THE USE,
 *     REPRODUCTION, MODIFICATION OR DISTRIBUTION OF THE SOFTWARE OR
 *     DOCUMENTATION WILL NOT INFRINGE A THIRD PARTY'S INTELLECTUAL
 *     PROPERTY RIGHTS. HP DOES NOT WARRANT THAT THE SOFTWARE OR
 *     DOCUMENTATION IS ERROR FREE. HP DISCLAIMS ALL WARRANTIES,
 *     EXPRESS AND IMPLIED, WITH REGARD TO THE SOFTWARE AND THE
 *     DOCUMENTATION. HP SPECIFICALLY DISCLAIMS ALL WARRANTIES OF
 *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * 5.  HEWLETT-PACKARD COMPANY WILL NOT IN ANY EVENT BE LIABLE FOR ANY
 *     DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
 *     (INCLUDING LOST PROFITS) RELATED TO ANY USE, REPRODUCTION,
 *     MODIFICATION, OR DISTRIBUTION OF THE SOFTWARE OR DOCUMENTATION.
 *
 * dn - The utility program that reads or changes the dummynet's
 *      default settings.
 *
 * All the following options require an option argument.
 * 
 * General options
 *    -p  Push dummynet between 0: IP and driver, 1: TCP and IP
 * Outbound options
 *    -e  Enable/disable simulation - 1: enable, 0: disable
 *    -b  Bandwidth in bps
 *    -d  Delay (pipe latency) in ticks
 *    -q  Queue size in # of packets
 *    -s  queue Size in bytes
 *    -f  Frame payload size
 *    -n  Number of frame errors per a million frames
 *    -c  Cost of a local link-level frame retransmission
 *    -t  # of Trials of local rexmits excluding the original transmit
 * Inbound options
 *    -E  Enable/disable simulation - 1: enable, 0: disable
 *    -B  Bandwidth in bps
 *    -D  Delay (pipe latency) in ticks
 *    -Q  Queue size in # of packets
 *    -S  queue Size in bytes
 *    -F  Frame payload size
 *    -N  Number of frame errors per a million frames
 *    -C  Cost of a local link-level frame retransmission
 *    -T  # of Trials of local rexmits excluding the original transmit
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <varargs.h>
#include <time.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/strstat.h>
#include <sys/strlog.h>
#include <sys/sad.h>
#include <errno.h>
#include "dummynetopt.h"

/*
 * Functions
 */
void print_usage();
void show_current_defaults(int fd);
void show_option(dn_option_t *);
void set_dummynet_defaults(int);
void process_options(int, char **);

/*
 * Globals
 */
int dn_placement = -1;

int dn_outbound_enable = -1;
int dn_outbound_bandwidth_bps=-1;     /* bits per second                    */
int dn_outbound_delay_ticks=-1;       /* ticks                              */
int dn_outbound_queue_size=-1;        /* # of packets: 1st choice           */
int dn_outbound_queue_size_bytes=-1;  /* bytes:        2nd choice           */
int dn_outbound_frame_payload_size=-1;/* frame's payload size (bytes)       */
int dn_outbound_ferror_per_million=-1;/* # frame errors / one million frames*/
int dn_outbound_ferror_delay_ticks=-1;/* cost of a local retransmit in ticks*/
int dn_outbound_ferror_rxmit_trials=-1;/* # of attempts of local retransmits 
				       * excluding the original transmit    */
int dn_inbound_enable=-1;
int dn_inbound_bandwidth_bps=-1;      /* bits per second                    */
int dn_inbound_delay_ticks=-1;        /* ticks                              */
int dn_inbound_queue_size=-1;         /* # of packets: 1st choice           */
int dn_inbound_queue_size_bytes=-1;   /* bytes:        2nd choice           */
int dn_inbound_frame_payload_size=-1; /* frame's payload size (bytes)       */
int dn_inbound_ferror_per_million=-1; /* # frame errors / one million frames*/
int dn_inbound_ferror_delay_ticks=-1; /* cost of a local retransmit in ticks*/
int dn_inbound_ferror_rxmit_trials=-1;/* # of attempts of local retransmits 
				       * excluding the original transmit    */
void
print_usage()
{
    int		tick;
    
    tick = 1000/HZ;
    
    fprintf(stderr,"usage:\n");
    fprintf(stderr,"To read the current default settings:\n");
    fprintf(stderr,"  dn\n");
    fprintf(stderr,"To change the current default setting(s):\n");
    fprintf(stderr,"  dn <the following options>\n");
    fprintf(stderr,"    General options\n");
    fprintf(stderr,"     -p <0|1>    you push dummynet 0: between IP and driver\n");
    fprintf(stderr,"                                   1: between TCP and IP\n");
    fprintf(stderr,"    Outbound options\n");
    fprintf(stderr,"     -e <0|1>    enable (1) or disable (0) dummynet\n");
    fprintf(stderr,"     -b <bps>    bandwidth in bps\n");
    fprintf(stderr,"     -d <ticks>  delay (pipe latency) in ticks (%d ms)\n", tick);
    fprintf(stderr,"     -q <pkts>   queue size in # of packets\n");
    fprintf(stderr,"     -s (bytes>  queue size in # of bytes\n");
    fprintf(stderr,"     -f <bytes>  frame payload size\n");
    fprintf(stderr,"     -n <frames> # of frame errors per a million frames\n");
    fprintf(stderr,"     -c <ticks>  cost of a local frame rexmit (tick: %d ms)\n", tick);
    fprintf(stderr,"     -t <trials> # trials of local rexmits excluding original xmit\n");
    fprintf(stderr,"    Inbound options\n");
    fprintf(stderr,"     -E <0|1>    enable (1) or disable (0) dummynet\n");
    fprintf(stderr,"     -B <bps>    bandwidth in bps\n");
    fprintf(stderr,"     -D <ticks>  delay (pipe latency) in ticks (%d ms)\n", tick);
    fprintf(stderr,"     -Q <pkts>   queue size in # of packets\n");
    fprintf(stderr,"     -S (bytes>  queue size in # of bytes\n");
    fprintf(stderr,"     -S <bytes>  frame payload size\n");
    fprintf(stderr,"     -N <frames> # of frame errors per a million frames\n");
    fprintf(stderr,"     -C <ticks>  cost of a local frame rexmit (tick: %d ms)\n", tick);
    fprintf(stderr,"     -T <trials> # trials of local rexmits excluding original xmit\n");
}

main(int argc, char **argv)
{
    int			fd;
    dn_option_t		dnopts[DN_NUMBER_OF_OPTIONS];
    struct strioctl	strioc;
    int			n;
    int			i;
    int			set_defaults;
    int			get_defaults;
    
    if (argc == 1)
    {
	get_defaults = 1;
	set_defaults = 0;
    }
    else
    {
	process_options(argc, argv);
	get_defaults = 0;
	set_defaults = 1;
    }
    /*
     * Open sad (streams admin driver).
     */
    fd = open("/dev/sad", O_RDWR);
    if (fd == -1)
    {
	fprintf(stderr, "Failed to open /dev/sad: errno=%d: %s\n",
		errno, strerror(errno));
	exit(1);
    }
    /*
     * Push dummynet.
     */
    if (ioctl(fd, I_PUSH, "dummynet") == -1)
    {
	fprintf(stderr, "Failed to push dummynet module: errno=%d: %s\n",
		errno, strerror(errno));
	exit(1);
    }
    if (set_defaults)
    {
	/*
	 * Set the dummynet's default setting(s).
	 */
	set_dummynet_defaults(fd);
    }
    if (get_defaults)
    {
	/*
	 * Get the current dummynet's default settings.
	 */
	show_current_defaults(fd);
    }
    close(fd);
}

void
set_dummynet_defaults(int fd)
{
    dn_option_t		dnopts[DN_NUMBER_OF_OPTIONS];
    struct strioctl	strioc;
    int			i = 0;
    int			r;

    /*
     * General options.
     */
    if (dn_placement != -1)
    {
	dnopts[i].name  = DN_PLACEMENT;
	dnopts[i].value = dn_placement;
	i++;
    }
    /*
     * Outbound options.
     */
    if (dn_outbound_enable != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_ENABLE;
	dnopts[i].value = dn_outbound_enable;
	i++;
    }
    if (dn_outbound_bandwidth_bps != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_BANDWIDTH_BPS;
	dnopts[i].value = dn_outbound_bandwidth_bps;
	i++;
    }
    if (dn_outbound_delay_ticks != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_DELAY_TICKS;
	dnopts[i].value = dn_outbound_delay_ticks;
	i++;
    }
    if (dn_outbound_queue_size != -1) 
    {
	dnopts[i].name  = DN_OUTBOUND_QUEUE_SIZE;
	dnopts[i].value = dn_outbound_queue_size;
	i++;
    }
    if (dn_outbound_queue_size_bytes != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_QUEUE_SIZE_BYTES;
	dnopts[i].value = dn_outbound_queue_size_bytes;
	i++;
    }
    if (dn_outbound_frame_payload_size != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_FRAME_PAYLOAD_SIZE;
	dnopts[i].value = dn_outbound_frame_payload_size;
	i++;
    }
    if (dn_outbound_ferror_per_million != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_FERROR_PER_MILLION;
	dnopts[i].value = dn_outbound_ferror_per_million;
	i++;
    }
    if (dn_outbound_ferror_delay_ticks != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_FERROR_DELAY_TICKS;
	dnopts[i].value = dn_outbound_ferror_delay_ticks;
	i++;
    }
    if (dn_outbound_ferror_rxmit_trials != -1)
    {
	dnopts[i].name  = DN_OUTBOUND_FERROR_RXMIT_TRIALS;
	dnopts[i].value = dn_outbound_ferror_rxmit_trials;
	i++;
    }
    /*
     * Inbound options.
     */
    if (dn_inbound_enable != -1)
    {
	dnopts[i].name  = DN_INBOUND_ENABLE;
	dnopts[i].value = dn_inbound_enable;
	i++;
    }
    if (dn_inbound_bandwidth_bps != -1)
    {
	dnopts[i].name  = DN_INBOUND_BANDWIDTH_BPS;
	dnopts[i].value = dn_inbound_bandwidth_bps;
	i++;
    }
    if (dn_inbound_delay_ticks != -1)
    {
	dnopts[i].name  = DN_INBOUND_DELAY_TICKS;
	dnopts[i].value = dn_inbound_delay_ticks;
	i++;
    }
    if (dn_inbound_queue_size != -1) 
    {
	dnopts[i].name  = DN_INBOUND_QUEUE_SIZE;
	dnopts[i].value = dn_inbound_queue_size;
	i++;
    }
    if (dn_inbound_queue_size_bytes != -1)
    {
	dnopts[i].name  = DN_INBOUND_QUEUE_SIZE_BYTES;
	dnopts[i].value = dn_inbound_queue_size_bytes;
	i++;
    }
    if (dn_inbound_frame_payload_size != -1)
    {
	dnopts[i].name  = DN_INBOUND_FRAME_PAYLOAD_SIZE;
	dnopts[i].value = dn_inbound_frame_payload_size;
	i++;
    }
    if (dn_inbound_ferror_per_million != -1)
    {
	dnopts[i].name  = DN_INBOUND_FERROR_PER_MILLION;
	dnopts[i].value = dn_inbound_ferror_per_million;
	i++;
    }
    if (dn_inbound_ferror_delay_ticks != -1)
    {
	dnopts[i].name  = DN_INBOUND_FERROR_DELAY_TICKS;
	dnopts[i].value = dn_inbound_ferror_delay_ticks;
	i++;
    }
    if (dn_inbound_ferror_rxmit_trials != -1)
    {
	dnopts[i].name  = DN_INBOUND_FERROR_RXMIT_TRIALS;
	dnopts[i].value = dn_inbound_ferror_rxmit_trials;
	i++;
    }
    /*
     * Call ioctl.
     */
    strioc.ic_cmd    = DN_SET_DEFAULTS;
    strioc.ic_timout = 0;
    strioc.ic_len    = sizeof(dn_option_t) * i;
    strioc.ic_dp     = (char *)&dnopts[0];
    r = ioctl(fd, I_STR, &strioc);
    if (r == -1)
    {
	fprintf(stderr, "DN_SET_DEFAULTS ioctl failed: errno=%d: %s\n",
		errno, strerror(errno));
	exit(1);
    }
    else if (r != i)
    {
	fprintf(stderr,
		"Warning: DN_SET_DEFAULTS ioctl ignored %d option(s)\n",
		i - r);
    }
}

void
show_current_defaults(int fd)
{
    dn_option_t		dnopts[DN_NUMBER_OF_OPTIONS];
    struct strioctl	strioc;
    int			n;
    int			i;

    /*
     * Call ioctl.
     */
    strioc.ic_cmd    = DN_GET_ALL_DEFAULTS;
    strioc.ic_timout = 0;
    strioc.ic_len    = sizeof(dnopts);
    strioc.ic_dp     = (char *)&dnopts[0];
    if (ioctl(fd, I_STR, &strioc) == -1)
    {
	fprintf(stderr, "DN_GET_ALL_DEFAULTS ioctl failed: errno=%d: %s\n",
		errno, strerror(errno));
	exit(1);
    }
    /*
     * Display the settings.
     */
    printf("Default Option                  Value  Description\n");
    printf("------------------------------- ------ -----------\n");
    n = strioc.ic_len / sizeof(dn_option_t);
    for (i = 0; i < n; i++)
    {
	show_option(&dnopts[i]);
    }
    printf("(Note: tick = %d msec)\n", 1000/HZ);
}

void
show_option(dn_option_t *dnopt)
{
    char	*optname;
    int		optvalue;
    char	*description;

    optvalue = dnopt->value;
    switch (dnopt->name)
    {
	/*
	 * General options.
	 */
	case DN_PLACEMENT:
	    optname = "dn_placement";
	    switch (optvalue)
	    {
		case PUSH_DN_BETWEEN_IP_AND_DRIVER:
		    description = "You push dummynet between IP and driver.";
		    break;
		case PUSH_DN_BETWEEN_TCP_AND_IP:
		    description = "You push dummynet between TCP and IP.";
		    break;
		default:
		    description = "Invalid value";
		    break;
	    }
	    break;
        /*
	 * Outbound options.
	 */
	case DN_OUTBOUND_ENABLE:
	    optname     = "dn_outbound_enable";
	    if (optvalue > 0)
	    {
		description = "Outbound simulation ON";
	    }
	    else
	    {
		description = "Outbound simulation OFF";
	    }
	    break;
	case DN_OUTBOUND_BANDWIDTH_BPS:
	    optname     = "dn_outbound_bandwidth_bps";
	    description = "Outbound bandwidth (bps)";
	    break;
	case DN_OUTBOUND_DELAY_TICKS:
	    optname     = "dn_outbound_delay_ticks";
	    description = "Outbound delay in ticks";
	    break;
	case DN_OUTBOUND_QUEUE_SIZE:
	    optname     = "dn_outbound_queue_size";
	    if (optvalue == 0)
	    {
		description = "Outbound queue: no limit on # packets";
	    }
	    else
	    {
		description = "Outbound queue size in packets";
	    }
	    break;
	case DN_OUTBOUND_QUEUE_SIZE_BYTES:
	    optname     = "dn_outbound_queue_size_bytes";
	    if (optvalue == 0)
	    {
		description = "Outbound queue: no limit on # bytes";
	    }
	    else
	    {
		description = "Outbound queue size in bytes";
	    }
	    break;
	case DN_OUTBOUND_FRAME_PAYLOAD_SIZE:
	    optname     = "dn_outbound_frame_payload_size";
	    if (optvalue == 0)
	    {
		description = "No frame error simulation.";
	    }
	    else
	    {
		description = "Outbound frame payload size in bytes";
	    }
	    break;
	case DN_OUTBOUND_FERROR_PER_MILLION:
	    optname     = "dn_outbound_ferror_per_million";
	    description = "Outbound # frame errors per million";
	    break;
	case DN_OUTBOUND_FERROR_DELAY_TICKS:
	    optname     = "dn_outbound_ferror_delay_ticks";
	    description = "Outbound frame rexmit penalty";
	    break;
	case DN_OUTBOUND_FERROR_RXMIT_TRIALS:
	    optname     = "dn_outbound_ferror_rxmit_trials";
	    description = "Outbound # trials of link-level rexmits";
	    break;
        /*
	 * Inbound options.
	 */
	case DN_INBOUND_ENABLE:
	    optname     = "dn_inbound_enable";
	    if (optvalue > 0)
	    {
		description = "Inbound simulation ON";
	    }
	    else
	    {
		description = "Inbound simulation OFF";
	    }
	    break;
	case DN_INBOUND_BANDWIDTH_BPS:
	    optname     = "dn_inbound_bandwidth_bps";
	    description = "Inbound bandwidth (bps)";
	    break;
	case DN_INBOUND_DELAY_TICKS:
	    optname     = "dn_inbound_delay_ticks";
	    description = "Inbound delay in ticks";
	    break;
	case DN_INBOUND_QUEUE_SIZE:
	    optname     = "dn_inbound_queue_size";
	    if (optvalue == 0)
	    {
		description = "Inbound queue: no limit on # packets";
	    }
	    else
	    {
		description = "Inbound queue size in packets";
	    }
	    break;
	case DN_INBOUND_QUEUE_SIZE_BYTES:
	    optname     = "dn_inbound_queue_size_bytes";
	    if (optvalue == 0)
	    {
		description = "Inbound queue: no limit on # bytes";
	    }
	    else
	    {
		description = "Inbound queue size in bytes";
	    }
	    break;
	case DN_INBOUND_FRAME_PAYLOAD_SIZE:
	    optname     = "dn_inbound_frame_payload_size";
	    if (optvalue == 0)
	    {
		description = "No frame error simulation.";
	    }
	    else
	    {
		description = "Inbound frame payload size in bytes";
	    }
	    break;
	case DN_INBOUND_FERROR_PER_MILLION:
	    optname     = "dn_inbound_ferror_per_million";
	    description = "Inbound # frame errors per million";
	    break;
	case DN_INBOUND_FERROR_DELAY_TICKS:
	    optname     = "dn_inbound_ferror_delay_ticks";
	    description = "Inbound frame rexmit penalty";
	    break;
	case DN_INBOUND_FERROR_RXMIT_TRIALS:
	    optname     = "dn_inbound_ferror_rxmit_trials";
	    description = "Inbound # trials of link-level rexmits";
	    break;

	default:
	    optname = "Invalid option";
	    description = "";
	    break;
    }
    printf("%-31s %6d %s\n", optname, optvalue, description);
}

void
process_options(int argc, char **argv)
{
    int		c;
    int		errflg = 0;

    extern char	*optarg;
    extern int	optopt;
    
    while ((c = getopt(argc, argv, ":p:e:b:d:q:s:f:n:c:t:E:B:D:Q:S:F:N:C:T:"))
	   != -1)
    {
	switch (c)
	{
	    /*
	     * General options.
	     */
	    case 'p':
		dn_placement = atoi(optarg);
		break;
	    /*
	     * Outbound options.
	     */
	    case 'e':
		dn_outbound_enable = atoi(optarg);
		break;
	    case 'b':
		dn_outbound_bandwidth_bps = atoi(optarg);
		break;
	    case 'd':
	        dn_outbound_delay_ticks = atoi(optarg);
		break;
	    case 'q':
		dn_outbound_queue_size = atoi(optarg);
		break;
	    case 's':
		dn_outbound_queue_size_bytes = atoi(optarg);
		break;
	    case 'f':
		dn_outbound_frame_payload_size = atoi(optarg);
		break;
	    case 'n':
		dn_outbound_ferror_per_million = atoi(optarg);
		break;
	    case 'c':
		dn_outbound_ferror_delay_ticks = atoi(optarg);
		break;
	    case 't':
		dn_outbound_ferror_rxmit_trials = atoi(optarg);
		break;
	    /*
	     * Inbound options.
	     */
	    case 'E':
		dn_inbound_enable = atoi(optarg);
		break;
	    case 'B':
		dn_inbound_bandwidth_bps = atoi(optarg);
		break;
	    case 'D':
	        dn_inbound_delay_ticks = atoi(optarg);
		break;
	    case 'Q':
		dn_inbound_queue_size = atoi(optarg);
		break;
	    case 'S':
		dn_inbound_queue_size_bytes = atoi(optarg);
		break;
	    case 'F':
		dn_inbound_frame_payload_size = atoi(optarg);
		break;
	    case 'N':
		dn_inbound_ferror_per_million = atoi(optarg);
		break;
	    case 'C':
		dn_inbound_ferror_delay_ticks = atoi(optarg);
		break;
	    case 'T':
		dn_inbound_ferror_rxmit_trials = atoi(optarg);
		break;

	    case ':':
		fprintf(stderr, "Option -%c requires an argument\n", optopt);
		errflg++;
		break;
	    case '?':
		fprintf(stderr, "Unrecognized optino: -%c\n", optopt);
		errflg++;
		break;
	}
    }
    if (errflg)
    {
	print_usage();
	exit (2);
    }
}

